@@ -11,14 +11,14 @@ import GUI.Types
 
 import Graphics.UI.Gtk
 
+import Control.Exception (assert)
+import Control.Monad
 import Data.Array
+import qualified Data.IntMap as IM
 import Data.IORef
+import Data.List as L
 import Data.Maybe
 import Data.Word (Word64)
-import Data.List as L
-import qualified Data.IntMap as IM
-import Control.Monad
-import Control.Exception (assert)
 import Numeric (showFFloat)
 import Text.Printf
 
@@ -104,11 +104,11 @@ summaryViewNew builder = do
     addGcColumn "Par collections" $ \(GcStatsEntry _ _ pcolls _ _ _) ->
       [ cellText := show pcolls ]
     addGcColumn "Elapsed time"    $ \(GcStatsEntry _ _ _ time _ _) ->
-      [ cellText := printf "%5.2fs" (timeToSecondsDbl time) ]
+      [ cellText := (printf "%5.2fs" (timeToSecondsDbl time) :: String) ]
     addGcColumn "Avg pause"       $ \(GcStatsEntry _ _ _ _ avgpause _) ->
-      [ cellText := printf "%3.4fs" avgpause ]
+      [ cellText := (printf "%3.4fs" avgpause :: String) ]
     addGcColumn "Max pause"       $ \(GcStatsEntry _ _ _ _ _ maxpause) ->
-      [ cellText := printf "%3.4fs" maxpause ]
+      [ cellText := (printf "%3.4fs" maxpause :: String) ]
 
     treeviewSparkStats <- getWidget castToTreeView "treeviewSparkStats"
     treeViewSetModel treeviewSparkStats storeSparkStats
@@ -266,7 +266,9 @@ ppWithCommas =
 
 setSummaryStatsEmpty :: SummaryView -> IO ()
 setSummaryStatsEmpty SummaryView{..} = do
-  mapM_ (\label -> set label [ labelText := "", widgetTooltipText := Nothing ]) $
+  mapM_ (\label -> set label [ labelText := ""
+                             , widgetTooltipText
+                               := (Nothing :: Maybe String) ]) $
     [ labelTimeTotal, labelTimeMutator
     , labelTimeGC, labelTimeProductivity ] ++
     [ w
@@ -281,7 +283,8 @@ setHeapStatsAvailable :: SummaryView -> Bool -> IO ()
 setHeapStatsAvailable SummaryView{..} available
   | available = do
       forM_ unavailableWidgets $ \widget ->
-        set widget [ widgetTooltipText := Nothing, widgetSensitive := True ]
+        set widget [ widgetTooltipText := (Nothing :: Maybe String)
+                   , widgetSensitive := True ]
 
   | otherwise = do
       forM_ allLabels $ \label -> set label [ labelText := "" ]
@@ -835,9 +838,38 @@ accumEvent !statsAccum (CapEvent mcap ev) =
                   ModeIdle -> dGC
                   -- Impossible.
                   ModeInit  -> error "scanEvents: GCStatsGHC ModeInit"
-                  ModeStart -> error "scanEvents: GCStatsGHC ModeStart"
                   ModeGHC{} -> error "scanEvents: GCStatsGHC ModeGHC"
-                  ModeEnd   -> error "scanEvents: GCStatsGHC ModeEnd"
+                  -- The last two cases are copied from case @GlobalSyncGC@
+                  -- to work around low-resolution timestamps (#35).
+                  -- Normally, these states would be impossible here, because
+                  -- @GlobalSyncGC@ would already transition away from these
+                  -- states. But if @GlobalSyncGC@ comes too early, the states
+                  -- can appear here. The computed stats are usually only
+                  -- slightly different than if @GlobalSyncGC@ made the state
+                  -- transitions, because the timestamps of @GCStatsGHC@
+                  -- and @GlobalSyncGC@ are normally only slightly different.
+                  --
+                  -- Cap takes part in the GC (not known if seq or par).
+                  -- Here is the moment where all caps taking place in the GC
+                  -- are identified and we can aggregate all their data
+                  -- at once (currently we just increment a counter for each).
+                  -- The EndGC events can come much later for some caps and at
+                  -- that time other caps are already inside their new GC.
+                  ModeStart ->
+                    dGC { gcMode = ModeSync cap
+                        , gcGenStat =
+                            if capKey == cap
+                            then IM.insert gcGenTot
+                                   totGC{ gcAll = gcAll totGC + 1 }
+                                   gcGenStat
+                            else gcGenStat
+                        }
+                  -- Cap is not in the GC. Mark it as idle to complete
+                  -- the identification of caps that take part
+                  -- in the current GC. Without overwritin the mode,
+                  -- the cap could be processed later on as if
+                  -- it took part in the GC, giving wrong results.
+                  ModeEnd  -> dGC { gcMode = ModeIdle }
           EndGC ->
             assert (gcMode capGC `notElem` [ModeEnd, ModeIdle]) $
             let endedGC = capGC { gcMode = ModeEnd }
@@ -892,4 +924,4 @@ accumEvent !statsAccum (CapEvent mcap ev) =
             in sd { dsparkTable =
                       IM.alter (alterCounter current) cap dsparkTable }
           _ -> sd
-    in scan (fromJust mcap) statsAccum ev
+    in scan (fromMaybe (error "Error: missing cap; use 'ghc-events validate' to verify the eventlog") mcap) statsAccum ev
